//
// Created by hangshu on 22-7-4.
//
#define DEBUG

#include "codegen.hh"
#include <set>
#include <algorithm>
#include "allocRegs.hh"

int GR::reg_num = 16;
int FR::reg_num = 32;
int bbNameCnt = 0;
std::map<std::string, std::string> bbNameMapping;
std::map<std::string, std::string> stringConstMapping;
int stringConstCnt = 0;


std::string Codegen::getFloatAddr(float x) {
    if (floatConstMapping.count(x) == 0) {
        floatConstMapping[x] = ".f" + std::to_string(floatConstCnt++);
    }
    return floatConstMapping[x];
}

bool is_legal_load_store_offset(int32_t offset) {
    return offset >= -4095 && offset <= 4095;
}

bool is_legal_immediate(int32_t value) {
    uint32_t u = static_cast<uint32_t>(value);
    if (u <= 0xffu) return true;
    for (int i = 1; i < 16; ++i) {
        uint32_t cur = (u << (2 * i)) | (u >> (32 - 2 * i));
        if (cur <= 0xffu) return true;
    }
    return false;
}

std::string getBBName(std::string name) {
    if (bbNameMapping.count(name) == 0) {
        bbNameMapping[name] = ".L" + std::to_string(bbNameCnt++);
    }
    return bbNameMapping[name];
}

std::vector<Instr *> setIntValue(GR target, uint32_t value) {
    if (0 <= value && value <= 65535) {
        return {new MovImm(target, value)};
    } else if (-255 <= value && value <= 0) {
        return {new MvnImm(target, -value - 1)};
    } else {
        int imm_low = value & ((1 << 16) - 1);
        int imm_high = value >> 16;
        return {new MoveW(target, imm_low),
                new MoveT(target, imm_high)};
    }
}

std::vector<Instr *> setFloatValue(GR target, float value) {
    union {
        float floatVal;
        uint32_t intVal;
    } data;
    data.floatVal = value;
    return setIntValue(target, data.intVal);
}

uint32_t floatToInt(float value) {
    union {
        float floatVal;
        uint32_t intVal;
    } data;
    data.floatVal = value;
    return data.intVal;
}

GR Codegen::getConstantGR(int x, BasicBlock *block, std::vector<Instr *> &vec) {
    if (!constantIntMapping[block].count(x)) {
        GR gr = GR::allocateReg();
        for (Instr *instr: setIntValue(gr, x)) {
            vec.push_back(instr);
        }
        constantIntMapping[block][x] = gr;
    }
    return constantIntMapping[block][x];
}

GR Codegen::getSymbolGR(std::string symbol, BasicBlock *block, std::vector<Instr *> &vec) {
    if (!symbolMapping[block].count(symbol)) {
        GR gr = GR::allocateReg();
        vec.push_back(new MoveWFromSymbol(gr, symbol));
        vec.push_back(new MoveTFromSymbol(gr, symbol));
        symbolMapping[block][symbol] = gr;
    }
    return symbolMapping[block][symbol];
}

FR Codegen::getFloatFR(float x, BasicBlock *block, std::vector<Instr *> &vec) {
    if (!constantFloatMapping[block].count(x)) {
        FR fr = FR::allocateReg();
        GR gr = GR::allocateReg();
        vec.push_back(new MoveWFromSymbol(gr, getFloatAddr(x)));
        vec.push_back(new MoveTFromSymbol(gr, getFloatAddr(x)));
        vec.push_back(new VLoad(fr, gr, 0));
        constantFloatMapping[block][x] = fr;
    }
    return constantFloatMapping[block][x];
}

void Codegen::generateFloatConst() {
    for (std::pair<float, std::string> pair: floatConstMapping) {
        out << ".align\n";
        out << pair.second << ":\n";
        out << "\t.4byte " << floatToInt(pair.first) << "\n";
    }
}

void Codegen::generateMemset() {
    out << ".memset:\n";
    out << "\tpush {r4}\n";
    out << "\tmov r2,#0\n";
    out << "\tmov r3,#0\n";
    out << "\tmov r4,#8\n";
    out << ".memset8:\n";
    out << "\tsub r1,r1,#8\n";
    out << "\tcmp r1,#0\n";
    out << "\tblt .memset4\n";
    out << "\tstrd r2,r3,[r0],r4\n";
    out << "\tbne .memset8\n";;
    out << "\tb .memset_end\n";
    out << ".memset4:\n";
    out << "\tstr r2,[r0],#4\n";
    out << ".memset_end:\n";
    out << "\tpop {r4}\n";
    out << "\tbx lr\n";
}

void Codegen::generateProgramCode() {
    out << ".arch armv7ve\n";
    out << ".arm\n";
    out << ".fpu neon\n";
    out << ".text\n";
    out << ".global main\n";
    generateGlobalCode();
    out << ".section .text\n";
    generateMemset();
    int removeCnt = 0;
    Function *entry_func = new Function(".init", new Type(TypeID::VOID));
    entry_func->pushBB(irVisitor.entry);
    irVisitor.functions.push_back(entry_func);
    std::map<Function *, std::set<GR>> usedGRMapping;
    std::map<Function *, std::set<FR>> usedFRMapping;
    std::map<Function *, int> spillCountMapping;
    std::map<Function *, int> surplyFor8AlignMapping;
    for (auto itt = irVisitor.functions.rbegin(); itt != irVisitor.functions.rend(); itt++) {
        Function *function = *itt;
        if (function->basicBlocks.empty()) continue;
        //cope with phi
        {
            for (BasicBlock *block: function->basicBlocks) {
                for (Instruction *ir: block->getIr()) {
                    if (typeid(*ir) == typeid(PhiIR)) {
                        PhiIR *phiIr = dynamic_cast<PhiIR *>(ir);
                        for (std::pair<BasicBlock *, Value *> pair: phiIr->params) {
                            BasicBlock *pre = pair.first;
                            Value *src = pair.second;
                            if (src == nullptr) continue;
                            if (src == phiIr->dst) {
                                continue;
                            }
                            if (typeid(*src) == typeid(TempVal)) {
                                TempVal *sr = dynamic_cast<TempVal *>(src);
                                if (sr->getVal() == phiIr->dst) {
                                    continue;
                                }
                            }
                            if (pre->getIr().empty()) {
                                pre->getIr().push_back(new MoveIR(phiIr->dst, src));
                                continue;
                            }
                            auto it = pre->getIr().end() - 1;
                            if (typeid(**it) == typeid(JumpIR)) {
                                pre->getIr().insert(it, new MoveIR(phiIr->dst, src));
                            } else if (typeid(**it) == typeid(BranchIR)) {
                                Instruction *p = *(it - 1);
                                if (it != pre->getIr().begin() && (typeid(*p) == typeid(GTIIR)
                                                                   || typeid(*p) != typeid(GTFIR) ||
                                                                   typeid(*p) != typeid(LTIIR)
                                                                   || typeid(*p) != typeid(LTFIR) ||
                                                                   typeid(*p) != typeid(GEIIR)
                                                                   || typeid(*p) != typeid(GEFIR) ||
                                                                   typeid(*p) != typeid(LEIIR)
                                                                   || typeid(*p) != typeid(LEFIR) ||
                                                                   typeid(*p) != typeid(EQUIIR) ||
                                                                   typeid(*p) != typeid(EQUFIR) ||
                                                                   typeid(*p) != typeid(NEIIR) ||
                                                                   typeid(*p) != typeid(NEFIR)))
                                    pre->getIr().insert(it - 1, new MoveIR(phiIr->dst, src));
                                else {
                                    pre->getIr().insert(it, new MoveIR(phiIr->dst, src));
                                }
                            } else {
                                pre->getIr().push_back(new MoveIR(phiIr->dst, src));
                            }
                        }
                    }
                }
            }
        }
        function->stackSize = translateFunction(function);
        if (function->name == ".init") {
            function->basicBlocks[0]->pushInstr(new Ret());
        }
        ColoringAlloc coloringAlloc(function);
        int spill_size = coloringAlloc.run() * 4;
        spillCountMapping[function] = spill_size;
        function->stackSize += spill_size;
        std::set<GR> allUsedRegsGR{GR(14)};
        std::set<FR> allUsedRegsFR;
        std::set<GR> callerSave;
        for (BasicBlock *block: function->basicBlocks) {
            for (auto it = block->getInstrs().rbegin(); it != block->getInstrs().rend();) {
                Instr *instr = *it;
                instr->replace(coloringAlloc.getColorGR(), coloringAlloc.getColorFR());
                instr->replaceBBName(bbNameMapping);
                if (typeid(*instr) == typeid(MoveReg) && instr->getUseG()[0] == instr->getDefG()[0] &&
                    (dynamic_cast<MoveReg *>(instr)->asr == -1) ||
                    typeid(*instr) == typeid(VMoveReg) && instr->getUseF()[0] == instr->getDefF()[0]) {
                    block->getInstrs().erase((++it).base());
                    removeCnt++;
                } else {
                    for (GR gr: instr->getDefG()) {
                        if (callee_save_regs.count(gr) != 0) {
                            allUsedRegsGR.insert(gr);
                        }
                        if (caller_save_regs.count(gr) != 0) {
                            callerSave.erase(gr);
                        }
                    }
                    for (GR gr: instr->getUseG()) {
                        if (caller_save_regs.count(gr) != 0) {
                            callerSave.insert(gr);
                        }
                    }
                    for (FR fr: instr->getDefF()) {
                        allUsedRegsFR.insert(fr);
                    }
                    if (typeid(*instr) == typeid(Bl)) {
//                        Push* pushInstr = dynamic_cast<Push*>(*(it+1));
//                        Pop* popInstr = dynamic_cast<Pop*>(*(it-1));
//                        if (pushInstr && popInstr) {
//                            pushInstr->addRegs(callerSave);
//                            popInstr->addRegs(callerSave);
//                        }
                    }
                    it++;
                }
            }
        }
        usedGRMapping[function] = allUsedRegsGR;
        allUsedRegsFR.erase(FR(0));
        usedFRMapping[function] = allUsedRegsFR;
        //make sp%8 == 0
        if ((function->stackSize + spill_size + allUsedRegsGR.size() * 4 + allUsedRegsFR.size() * 4) % 8 != 0) {
            surplyFor8AlignMapping[function] = 4;
            function->stackSize += 4;
        } else {
            surplyFor8AlignMapping[function] = 0;
        }
    }
//    std::cout << "after delete phi:\n";
//    irVisitor.print();
    for (auto itt = irVisitor.functions.rbegin(); itt != irVisitor.functions.rend(); itt++) {
        Function *function = *itt;
        if (function->basicBlocks.empty()) continue;
        function->basicBlocks.insert(function->basicBlocks.begin(), new NormalBlock(function->name));
        bbNameMapping[function->name] = function->name;
        if (function->name == "main") {
            function->basicBlocks[0]->getInstrs().insert(function->basicBlocks[0]->getInstrs().begin(),
                                                         new Bl(".init"));
        }
        //sub r11,sp, #size;
//            if (spillCountMapping[function] != 0) {
//                if (is_legal_load_store_offset(spillCountMapping[function]) &&
//                    is_legal_immediate(spillCountMapping[function])) {
//                    function->basicBlocks[0]->getInstrs().insert(function->basicBlocks[0]->getInstrs().begin(),
//                                                                 new GRegImmInstr(GRegImmInstr::Sub,GR(11),GR(13),spillCountMapping[function]));
//                } else {
//                    function->basicBlocks[0]->getInstrs().insert(function->basicBlocks[0]->getInstrs().begin(),
//                                                                 new GRegRegInstr(GRegRegInstr::Sub,GR(11),GR(13),GR(12)));
//                    std::vector<Instr *> vec = setIntValue(GR(12), spillCountMapping[function]);
//                    for (int i = vec.size() - 1; i >= 0; --i) {
//                        function->basicBlocks[0]->getInstrs().insert(function->basicBlocks[0]->getInstrs().begin(), vec[i]);
//                    }
//                }
//            }
        if (is_legal_immediate(function->stackSize) && is_legal_load_store_offset(function->stackSize)) {
            function->basicBlocks[0]->getInstrs().insert(function->basicBlocks[0]->getInstrs().begin(),
                                                         new GRegImmInstr(GRegImmInstr::Sub, GR(13), GR(13),
                                                                          function->stackSize));
        } else {
            function->basicBlocks[0]->getInstrs().insert(function->basicBlocks[0]->getInstrs().begin(),
                                                         new GRegRegInstr(GRegRegInstr::Sub, GR(13), GR(13),
                                                                          GR(12)));
            std::vector<Instr *> vec = setIntValue(GR(12), function->stackSize);
            for (int i = vec.size() - 1; i >= 0; --i) {
                function->basicBlocks[0]->getInstrs().insert(function->basicBlocks[0]->getInstrs().begin(), vec[i]);
            }
            usedGRMapping[function].insert(GR(12));
        }
        if (usedFRMapping[function].size() <= 16) {
            function->basicBlocks[0]->getInstrs().insert(function->basicBlocks[0]->getInstrs().begin(),
                                                         new Vpush(usedFRMapping[function]));
        } else {
            std::set<FR> first, second;
            int cnt = 0;
            for (auto item: usedFRMapping[function]) {
                if (cnt < 16) {
                    first.insert(item);
                } else {
                    second.insert(item);
                }
                cnt++;
            }
            function->basicBlocks[0]->getInstrs().insert(function->basicBlocks[0]->getInstrs().begin(),
                                                         new Vpush(first));
            function->basicBlocks[0]->getInstrs().insert(function->basicBlocks[0]->getInstrs().begin(),
                                                         new Vpush(second));
        }
        function->basicBlocks[0]->getInstrs().insert(function->basicBlocks[0]->getInstrs().begin(),
                                                     new Push(usedGRMapping[function]));
        //simple way

        for (BasicBlock *block: function->basicBlocks) {
            for (auto it = block->getInstrs().begin(); it != block->getInstrs().end();) {
                Instr *instr = *it;
                if (typeid(*instr) == typeid(Load)) {
                    Load *load = dynamic_cast<Load *>(instr);
                    if (load->offset < 0) {
                        load->offset = -load->offset + usedGRMapping[function].size() * 4 +
                                       usedFRMapping[function].size() * 4 + spillCountMapping[function] +
                                       surplyFor8AlignMapping[function];
                    }
                    if (!is_legal_immediate(load->offset) || !is_legal_load_store_offset(load->offset)) {
                        it = block->getInstrs().erase(it);
                        it = block->getInstrs().insert(it, new Load(load->dst, GR(14), 0));
                        it = block->getInstrs().insert(it,
                                                       new GRegRegInstr(GRegRegInstr::Add, GR(14), GR(14), load->base));
                        std::vector<Instr *> vv = setIntValue(GR(14), load->offset);
                        for (auto item = vv.rbegin(); item != vv.rend(); item++) {
                            it = block->getInstrs().insert(it, *item);
                        }
                        it = it + 1 + vv.size();
                    }
                }
                if (typeid(*instr) == typeid(Store)) {
                    Store *store = dynamic_cast<Store *>(instr);
                    if (store->offset < 0) {
                        store->offset = -store->offset + usedGRMapping[function].size() * 4 +
                                        usedFRMapping[function].size() * 4 + spillCountMapping[function] +
                                        surplyFor8AlignMapping[function];
                    }
                    if (!is_legal_immediate(store->offset) || !is_legal_load_store_offset(store->offset)) {
                        it = block->getInstrs().erase(it);
                        it = block->getInstrs().insert(it, new Store(store->src, GR(14), 0));
                        it = block->getInstrs().insert(it, new GRegRegInstr(GRegRegInstr::Add, GR(14), GR(14),
                                                                            store->base));
                        std::vector<Instr *> vv = setIntValue(GR(14), store->offset);
                        for (auto item = vv.rbegin(); item != vv.rend(); item++) {
                            it = block->getInstrs().insert(it, *item);
                        }
                        it = it + 1 + vv.size();
                    }
                }
                if (typeid(*instr) == typeid(VLoad)) {
                    VLoad *vload = dynamic_cast<VLoad *>(instr);
                    if (vload->offset < 0) {
                        vload->offset = -vload->offset + usedGRMapping[function].size() * 4 +
                                        usedFRMapping[function].size() * 4 + spillCountMapping[function] +
                                        surplyFor8AlignMapping[function];
                    }
                    if (!is_legal_immediate(vload->offset) || !is_legal_load_store_offset(vload->offset)) {
                        it = block->getInstrs().erase(it);
                        it = block->getInstrs().insert(it, new VLoad(vload->dst, GR(14), 0));
                        it = block->getInstrs().insert(it, new GRegRegInstr(GRegRegInstr::Add, GR(14), GR(14),
                                                                            vload->base));
                        std::vector<Instr *> vv = setIntValue(GR(14), vload->offset);
                        for (auto item = vv.rbegin(); item != vv.rend(); item++) {
                            it = block->getInstrs().insert(it, *item);
                        }
                        it = it + 1 + vv.size();
                    }
                }
                if (typeid(*instr) == typeid(VStore)) {
                    VStore *vstore = dynamic_cast<VStore *>(instr);
                    if (vstore->offset < 0) {
                        vstore->offset = -vstore->offset + usedGRMapping[function].size() * 4 +
                                         usedFRMapping[function].size() * 4 + spillCountMapping[function] +
                                         surplyFor8AlignMapping[function];
                    }
                    if (!is_legal_immediate(vstore->offset) || !is_legal_load_store_offset(vstore->offset)) {
                        it = block->getInstrs().erase(it);
                        it = block->getInstrs().insert(it, new VStore(vstore->src, GR(14), 0));
                        it = block->getInstrs().insert(it,
                                                       new GRegRegInstr(GRegRegInstr::Add, GR(14), GR(14),
                                                                        vstore->base));
                        std::vector<Instr *> vv = setIntValue(GR(14), vstore->offset);
                        for (auto item = vv.rbegin(); item != vv.rend(); item++) {
                            it = block->getInstrs().insert(it, *item);
                        }
                        it = it + 1 + vv.size();
                    }
                }
//                if (typeid(*instr) == typeid(MoveReg)) {
//                    MoveReg* moveReg = dynamic_cast<MoveReg*>(instr);
//                    //mov sp,r11;
//                    if (moveReg->getUseG()[0] == GR(11)) {
//                        block->getInstrs().erase(it);
//                        int cnt = 0;
//                        if (spillCountMapping[function] != 0) {
//                            if (is_legal_immediate(spillCountMapping[function]) && is_legal_load_store_offset(spillCountMapping[function])) {
//                                block->getInstrs().insert(it,new GRegImmInstr(GRegImmInstr::Sub, GR(13),GR(13),spillCountMapping[function]));
//                                cnt++;
//                            } else {
//                                block->getInstrs().insert(it,new GRegRegInstr(GRegRegInstr::Sub,GR(13),GR(13),GR(12)));
//                                std::vector<Instr*> vv = setIntValue(GR(12), spillCountMapping[function]);
//                                for (auto item = vv.rbegin();item != vv.rend();it++) {
//                                    block->getInstrs().insert(it,*item);
//                                }
//                                cnt = cnt + 1 + vv.size();
//                            }
//                        }
//                        it = it + cnt - 1;
//                    }
//                    if (moveReg->getDefG()[0] == GR(11)) {
//                        block->getInstrs().erase(it);
//                        int cnt = 0;
//                        if (spillCountMapping[function] != 0) {
//                            if (is_legal_immediate(spillCountMapping[function]) && is_legal_load_store_offset(spillCountMapping[function])) {
//                                block->getInstrs().insert(it,new GRegImmInstr(GRegImmInstr::Add, GR(13),GR(13),spillCountMapping[function]));
//                                cnt++;
//                            } else {
//                                block->getInstrs().insert(it,new GRegRegInstr(GRegRegInstr::Add,GR(13),GR(13),GR(12)));
//                                std::vector<Instr*> vv = setIntValue(GR(12), spillCountMapping[function]);
//                                for (auto item = vv.rbegin();item != vv.rend();it++) {
//                                    block->getInstrs().insert(it,*item);
//                                }
//                                cnt = cnt + 1 + vv.size();
//                            }
//                        }
//                        it = it + cnt - 1;
//                    }
//                }

                if (typeid(*instr) == typeid(Ret)) {
                    block->getInstrs().erase(it);
                    if (is_legal_immediate(function->stackSize) && is_legal_load_store_offset(function->stackSize)) {
                        if (function->stackSize != 0)
                            block->getInstrs().push_back(
                                    new GRegImmInstr(GRegImmInstr::Add, GR(13), GR(13), function->stackSize));
                    } else {
                        std::vector<Instr *> vec = setIntValue(GR(12), function->stackSize);
                        for (int i = 0; i < vec.size(); i++) {
                            block->getInstrs().push_back(vec[i]);
                        }
                        block->getInstrs().push_back(new GRegRegInstr(GRegRegInstr::Add, GR(13), GR(13), GR(12)));
                    }
                    std::set<GR> setGR = usedGRMapping[function];
                    setGR.erase(GR(14));
                    setGR.insert(GR(15));
                    if (!usedFRMapping[function].empty()) {
                        if (usedFRMapping[function].size() <= 16) {
                            block->getInstrs().push_back(new Vpop(usedFRMapping[function]));
                        } else {
                            std::set<FR> first, second;
                            int cnt = 0;
                            for (auto item: usedFRMapping[function]) {
                                if (cnt < 16) {
                                    first.insert(item);
                                } else {
                                    second.insert(item);
                                }
                                cnt++;
                            }
                            if (!first.empty()) {
                                block->getInstrs().push_back(new Vpop(first));
                            }
                            if (!second.empty()) {
                                block->getInstrs().push_back(new Vpop(second));
                            }
                        }
                    }
                    if (!setGR.empty()) {
                        block->getInstrs().push_back(new Pop(setGR));
                    }
                    break;
                } else if (typeid(*instr) == typeid(Push)) {
                    Push *pushInstr = dynamic_cast<Push *>(instr);
                    if (pushInstr->regs.empty()) {
                        block->getInstrs().erase(it);
                    } else {
                        it++;
                    }
                } else if (typeid(*instr) == typeid(Pop)) {
                    Pop *pushInstr = dynamic_cast<Pop *>(instr);
                    if (pushInstr->regs.empty()) {
                        block->getInstrs().erase(it);
                    } else {
                        it++;
                    }
                } else if (typeid(*instr) == typeid(Vpop)) {
                    Vpop *pushInstr = dynamic_cast<Vpop *>(instr);
                    if (pushInstr->regs.empty()) {
                        block->getInstrs().erase(it);
                    } else {
                        it++;
                    }
                } else if (typeid(*instr) == typeid(Vpush)) {
                    Vpush *pushInstr = dynamic_cast<Vpush *>(instr);
                    if (pushInstr->regs.empty()) {
                        block->getInstrs().erase(it);
                    } else {
                        it++;
                    }
                } else if (typeid(*instr) == typeid(GRegImmInstr)) {
                    GRegImmInstr *g = dynamic_cast<GRegImmInstr *>(instr);
                    if ((g->op == GRegRegInstr::Add && g->src2 == 0 ||
                         g->op == GRegRegInstr::Sub && g->src2 == 0) && (g->dst == g->src1)) {
                        block->getInstrs().erase(it);
                    } else {
                        it++;
                    }
                } else {
                    it++;
                }
            }
        }
    }
    for (auto itt = irVisitor.functions.rbegin(); itt != irVisitor.functions.rend(); itt++) {
        Function *function = *itt;
        if (function->basicBlocks.empty()) continue;
        comment("spurly Size: " + std::to_string(surplyFor8AlignMapping[function]));
        comment("spilled Size: " + std::to_string(spillCountMapping[function]));
        comment("stack Size: " + std::to_string(function->stackSize));
        for (BasicBlock *block: function->basicBlocks) {
            out << getBBName(block->name) << ":\n";
            for (auto it = block->getInstrs().begin(); it != block->getInstrs().end();) {
                Instr *instr = *it;
                it++;
                out << "\t";
                instr->print(out);
            }
        }
    }
    generateFloatConst();
}

int Codegen::translateFunction(Function *function) {
    stackMapping.clear();
    int stackSize = 4;
    GR::reg_num = 16;
    FR::reg_num = 32;
    for (BasicBlock *bb: function->basicBlocks) {
        getBBName(bb->name);
        if (bb->ir.empty()) {
            continue;
        }
        for (Instruction *ir: bb->ir) {
            if (typeid(*ir) == typeid(CallIR)) {
                int size = 0;
                CallIR *callIr = dynamic_cast<CallIR *>(ir);
                int gr_cnt = 0;
                int fr_cnt = 0;
                for (TempVal v: callIr->args) {
                    if (!v.isFloat()) {
                        if (gr_cnt >= 4) {
                            size += 4;
                        }
                        gr_cnt++;
                    } else {
                        if (fr_cnt >= 32) {
                            size += 4;
                        }
                        fr_cnt++;
                    }
                }
                stackSize = stackSize > size ? stackSize : size;
            }
        }
    }
    for (BasicBlock *bb: function->basicBlocks) {
        for (Instruction *ir: bb->ir) {
            if (dynamic_cast<AllocIR *>(ir)) {
                AllocIR *allocIr = dynamic_cast<AllocIR *>(ir);
                stackMapping[allocIr->v] = stackSize;
                stackSize += allocIr->arrayLen * 4;
            }
        }
    }
    {
        int gr_cnt = 0;
        int fr_cnt = 0;
        int cnt = 0;
        for (int i = 0; i < function->params.size(); i++) {
            if (!function->params[i]->getType()->isFloat()) {
                if (gr_cnt < 4) {
//                    gRegMapping[function->params[i]] = GR(gr_cnt);
                    function->basicBlocks[0]->getInstrs().push_back(
                            new MoveReg(getGR(function->params[i]), GR(gr_cnt)));
                } else {
                    //push {}  -> stackSize + cnt * 4 + pushSize
                    // eg :push {r4,r5} int f(int a,int b,int c,int d,int e)
                    // e -> [sp, # (16 + 8)]
                    stackMapping[function->params[i]] = -(stackSize + cnt * 4);
                    cnt++;
                    function->basicBlocks[0]->getInstrs().push_back(
                            new Load(getGR(function->params[i]), GR(13), stackMapping[function->params[i]]));
                }
                gr_cnt++;
            } else {
                if (fr_cnt < 32) {
                    //fRegMapping[function->params[i]] = FR(fr_cnt);
                    function->basicBlocks[0]->getInstrs().push_back(
                            new VMoveReg(getFR(function->params[i]), FR(fr_cnt)));
                } else {
                    stackMapping[function->params[i]] = -(stackSize + cnt * 4);
                    cnt++;
                    function->basicBlocks[0]->getInstrs().push_back(
                            new VLoad(getFR(function->params[i]), GR(13), stackMapping[function->params[i]]));
                }
                fr_cnt++;
            }
        }
    }
    for (BasicBlock *bb: function->basicBlocks) {
        for (auto it = bb->getIr().begin(); it != bb->getIr().end();it++) {
            Instruction* ir = *it;
            std::vector<Instr *> vec = translateInstr(ir, bb, it);
            for (int i = 0; i < vec.size(); ++i) {
                bb->pushInstr(vec[i]);
            }
        }
    }
    return stackSize;
}

std::vector<Instr *> Codegen::translateInstr(Instruction *ir, BasicBlock *block, std::vector<Instruction*>::iterator& iter) {
    if (typeid(*ir) == typeid(MoveIR)) {
        MoveIR *moveIr = dynamic_cast<MoveIR *>(ir);
        Value *dst = dynamic_cast<Value *>(moveIr->dst);
        TempVal *src = dynamic_cast<TempVal *>(moveIr->src);
        std::vector<Instr *> vec;
        assert(dst != nullptr);
        if (src != nullptr) {
            if (dst->getType()->isInt()) {
                if (!src->getVal()) {
                    if (is_legal_immediate(src->getInt())) {
                        vec.push_back(new MovImm(getGR(dst), src->getInt()));
                    } else {
                        for (Instr *instr: setIntValue(GR(12), src->getInt())) {
                            vec.push_back(instr);
                        }
                        vec.push_back(new MoveReg(getGR(dst), GR(12)));
                    }
                } else {
                    vec.push_back(new MoveReg(getGR(dst), getGR(src->getVal())));
                }
            } else {
                if (!src->getVal()) {
                    FR d = getFloatFR(src->getFloat(), block, vec);
                    vec.push_back(new VMoveReg(getFR(dst), d));
                } else {
                    vec.push_back(new VMoveReg(getFR(dst), getFR(src->getVal())));
                }
            }
        } else {
            VarValue *src2 = dynamic_cast<VarValue *>(moveIr->src);
            if (dst->getType()->isInt()) {
                vec.push_back(new MoveReg(getGR(dst), getGR(src2)));
            } else {
                vec.push_back(new VMoveReg(getFR(dst), getFR(src2)));
            }
        }
        return vec;
    }
    if (typeid(*ir) == typeid(AllocIIR)) {
        AllocIIR *allocIir = dynamic_cast<AllocIIR *>(ir);
        std::vector<Instr *> vec;
        if (allocIir->isArray) {
            if (is_legal_immediate(stackMapping[allocIir->v]) &&
                is_legal_load_store_offset(stackMapping[allocIir->v])) {
                vec.push_back(new GRegImmInstr(GRegImmInstr::Add, GR(0), GR(13), stackMapping[allocIir->v]));
            } else {
                for (Instr *instr: setIntValue(GR(12), stackMapping[allocIir->v])) {
                    vec.push_back(instr);
                }
                vec.push_back(new GRegRegInstr(GRegRegInstr::Add, GR(0), GR(13), GR(12)));
            }
            for (Instr *instr: setIntValue(GR(1), allocIir->arrayLen * 4)) {
                vec.push_back(instr);
            }
            vec.push_back(new Bl(".memset"));
        }
        return vec;
    }
    if (typeid(*ir) == typeid(AllocFIR)) {
        AllocFIR *allocIir = dynamic_cast<AllocFIR *>(ir);
        std::vector<Instr *> vec;
        if (allocIir->isArray) {
            if (is_legal_immediate(stackMapping[allocIir->v]) &&
                is_legal_load_store_offset(stackMapping[allocIir->v])) {
                vec.push_back(new GRegImmInstr(GRegImmInstr::Add, GR(0), GR(13), stackMapping[allocIir->v]));
            } else {
                for (Instr *instr: setIntValue(GR(12), stackMapping[allocIir->v])) {
                    vec.push_back(instr);
                }
                vec.push_back(new GRegRegInstr(GRegRegInstr::Add, GR(0), GR(13), GR(12)));
            }
            for (Instr *instr: setIntValue(GR(1), allocIir->arrayLen * 4)) {
                vec.push_back(instr);
            }
            vec.push_back(new Bl(".memset"));
        }
        return vec;
    }
    if (typeid(*ir) == typeid(GEPIR)) {
        GEPIR *gepIr = dynamic_cast<GEPIR *>(ir);
        std::vector<Instr *> vec;
        if (gepIr->v2->is_Global()) {

            vec.push_back(new MoveWFromSymbol(GR(14), gepIr->v2->getName()));
            vec.push_back(new MoveTFromSymbol(GR(14), gepIr->v2->getName()));
            GR base = GR(14);
            GR dst = getGR(gepIr->v1);
            if (!gepIr->v3) {
                if (is_legal_load_store_offset(4 * gepIr->arrayLen) && is_legal_immediate(4 * gepIr->arrayLen)) {
                    vec.push_back(new GRegImmInstr(GRegImmInstr::Add, dst, base, 4 * gepIr->arrayLen));
                } else {
                    for (Instr *instr: setIntValue(GR(12), 4 * gepIr->arrayLen)) {
                        vec.push_back(instr);
                    }
                    vec.push_back(new GRegRegInstr(GRegRegInstr::Add, dst, base, GR(12)));
                }
            } else {
                GR mul1 = getConstantGR(4, block, vec);
                vec.push_back(new MLA(dst, mul1, getGR(gepIr->v3), base));
            }
        } else {
            if (!gepIr->v3) {
                GR dst = getGR(gepIr->v1);
                if (stackMapping.count(gepIr->v2) != 0) {
                    if (is_legal_load_store_offset(gepIr->arrayLen * 4 + stackMapping[gepIr->v2]) &&
                        is_legal_immediate(4 * gepIr->arrayLen + stackMapping[gepIr->v2])) {
                        vec.push_back(new GRegImmInstr(GRegImmInstr::Add, dst, GR(13),
                                                       gepIr->arrayLen * 4 + stackMapping[gepIr->v2]));
                    } else {
                        for (Instr *instr: setIntValue(GR(12), gepIr->arrayLen * 4 + stackMapping[gepIr->v2])) {
                            vec.push_back(instr);
                        }
                        vec.push_back(new GRegRegInstr(GRegRegInstr::Add, dst, GR(13), GR(12)));
                    }
                } else {
                    if (is_legal_load_store_offset(gepIr->arrayLen * 4) && is_legal_immediate(4 * gepIr->arrayLen)) {
                        vec.push_back(new GRegImmInstr(GRegImmInstr::Add, dst, getGR(gepIr->v2), gepIr->arrayLen * 4));
                    } else {
                        for (Instr *instr: setIntValue(GR(12), gepIr->arrayLen * 4)) {
                            vec.push_back(instr);
                        }
                        vec.push_back(new GRegRegInstr(GRegRegInstr::Add, dst, getGR(gepIr->v2), GR(12)));
                    }
                }
            } else {
                //v3 * 4 + v2
                GR mul1 = getConstantGR(4, block, vec);
                GR dst = getGR(gepIr->v1);
                if (stackMapping.count(gepIr->v2) != 0) {
                    if (is_legal_load_store_offset(stackMapping[gepIr->v2]) &&
                        is_legal_immediate(stackMapping[gepIr->v2])) {
                        vec.push_back(
                                new GRegImmInstr(GRegImmInstr::Add, getGR(gepIr->v2), GR(13), stackMapping[gepIr->v2]));
                    } else {
                        for (Instr *instr: setIntValue(GR(12), stackMapping[gepIr->v2])) {
                            vec.push_back(instr);
                        }
                        vec.push_back(new GRegRegInstr(GRegRegInstr::Add, getGR(gepIr->v2), GR(13), GR(12)));
                    }
                    vec.push_back(new MLA(dst, mul1, getGR(gepIr->v3), getGR(gepIr->v2)));
                } else {
                    vec.push_back(new MLA(dst, mul1, getGR(gepIr->v3), getGR(gepIr->v2)));
                }
            }
        }
        return vec;
    }
    if (typeid(*ir) == typeid(LoadIIR)) {
        LoadIIR *loadIr = dynamic_cast<LoadIIR *>(ir);
        std::vector<Instr *> vec;
        GR dst = getGR(loadIr->v1);
        if (loadIr->v2->is_Global()) {
            vec.push_back(new MoveWFromSymbol(GR(12), loadIr->v2->getName()));
            vec.push_back(new MoveTFromSymbol(GR(12), loadIr->v2->getName()));
            GR base = GR(12);
            vec.push_back(new Load(dst, base, 0));
        } else {
            if (stackMapping.count(loadIr->v2) != 0) {
                if (is_legal_load_store_offset(stackMapping[loadIr->v2]) &&
                    is_legal_immediate(stackMapping[loadIr->v2])
                    || stackMapping[loadIr->v2] < 0) {
                    vec.push_back(new Load(dst, GR(13), stackMapping[loadIr->v2]));
                } else {
                    for (Instr *instr: setIntValue(GR(12), stackMapping[loadIr->v2])) {
                        vec.push_back(instr);
                    }
                    GR gr2 = GR::allocateReg();
                    vec.push_back(new GRegRegInstr(GRegRegInstr::Add, gr2, GR(12), GR(13)));
                    vec.push_back(new Load(dst, gr2, 0));
                }
            } else {
                vec.push_back(new Load(dst, getGR(loadIr->v2), 0));
            }
        }
        return vec;
    }
    if (typeid(*ir) == typeid(LoadFIR)) {
        LoadFIR *loadIr = dynamic_cast<LoadFIR *>(ir);
        std::vector<Instr *> vec;
        FR dst = getFR(loadIr->v1);
        if (loadIr->v2->is_Global()) {
            vec.push_back(new MoveWFromSymbol(GR(12), loadIr->v2->getName()));
            vec.push_back(new MoveTFromSymbol(GR(12), loadIr->v2->getName()));
            GR base = GR(12);
            vec.push_back(new VLoad(dst, base, 0));
        } else {
            if (stackMapping.count(loadIr->v2) != 0) {
                if (is_legal_load_store_offset(stackMapping[loadIr->v2]) &&
                    is_legal_immediate(stackMapping[loadIr->v2])
                    || stackMapping[loadIr->v2] < 0) {
                    vec.push_back(new VLoad(dst, GR(13), stackMapping[loadIr->v2]));
                } else {
                    for (Instr *instr: setIntValue(GR(12), stackMapping[loadIr->v2])) {
                        vec.push_back(instr);
                    }
                    GR gr2 = GR::allocateReg();
                    vec.push_back(new GRegRegInstr(GRegRegInstr::Add, gr2, GR(12), GR(13)));
                    vec.push_back(new VLoad(dst, gr2, 0));
                }
            } else {
                vec.push_back(new VLoad(dst, getGR(loadIr->v2), 0));
            }
        }
        return vec;
    }
    if (typeid(*ir) == typeid(StoreIIR)) {
        StoreIIR *storeIr = dynamic_cast<StoreIIR *>(ir);
        std::vector<Instr *> vec;
        GR src;
        if (!storeIr->src.getVal()) {
            for (Instr *instr: setIntValue(GR(14), storeIr->src.getInt())) {
                vec.push_back(instr);
            }
            src = GR(14);
        } else {
            if (gRegMapping.count(storeIr->src.getVal()) != 0) {
                src = getGR(storeIr->src.getVal());
            } else {
                vec.push_back(new Load(getGR(storeIr->src.getVal()), GR(13), stackMapping[storeIr->src.getVal()]));
                src = getGR(storeIr->src.getVal());
            }
        }
        if (storeIr->dst->is_Global()) {
            vec.push_back(new MoveWFromSymbol(GR(12), storeIr->dst->getName()));
            vec.push_back(new MoveTFromSymbol(GR(12), storeIr->dst->getName()));
            GR base = GR(12);
            vec.push_back(new Store(src, base, 0));
        } else {
            if (stackMapping.count(storeIr->dst) != 0) {
                if (is_legal_load_store_offset(stackMapping[storeIr->dst]) &&
                    is_legal_immediate(stackMapping[storeIr->dst])
                    || stackMapping[storeIr->dst] < 0) {
                    vec.push_back(new Store(src, GR(13), stackMapping[storeIr->dst]));
                } else {
                    for (Instr *instr: setIntValue(GR(12), stackMapping[storeIr->dst])) {
                        vec.push_back(instr);
                    }
                    GR gr2 = GR::allocateReg();
                    vec.push_back(new GRegRegInstr(GRegRegInstr::Add, gr2, GR(12), GR(13)));
                    vec.push_back(new Store(src, gr2, 0));
                }
            } else {
                vec.push_back(new Store(src, getGR(storeIr->dst), 0));
            }
        }
        return vec;
    }
    if (typeid(*ir) == typeid(StoreFIR)) {
        StoreFIR *storeIr = dynamic_cast<StoreFIR *>(ir);
        std::vector<Instr *> vec;
        FR src;
        if (!storeIr->src.getVal()) {
            src = getFloatFR(storeIr->src.getFloat(), block, vec);
        } else {
            if (fRegMapping.count(storeIr->src.getVal()) != 0) {
                src = getFR(storeIr->src.getVal());
            } else {
                vec.push_back(new Load(getGR(storeIr->src.getVal()), GR(13), stackMapping[storeIr->src.getVal()]));
                src = getFR(storeIr->src.getVal());
            }
        }
        if (storeIr->dst->is_Global()) {
            vec.push_back(new MoveWFromSymbol(GR(12), storeIr->dst->getName()));
            vec.push_back(new MoveTFromSymbol(GR(12), storeIr->dst->getName()));
            GR base = GR(12);
            vec.push_back(new VStore(src, base, 0));
        } else {
            if (stackMapping.count(storeIr->dst) != 0) {
                if (is_legal_load_store_offset(stackMapping[storeIr->dst]) &&
                    is_legal_immediate(stackMapping[storeIr->dst])
                    || stackMapping[storeIr->dst] < 0) {
                    vec.push_back(new VStore(src, GR(13), stackMapping[storeIr->dst]));
                } else {
                    for (Instr *instr: setIntValue(GR(12), stackMapping[storeIr->dst])) {
                        vec.push_back(instr);
                    }
                    GR gr2 = GR::allocateReg();
                    vec.push_back(new GRegRegInstr(GRegRegInstr::Add, gr2, GR(12), GR(13)));
                    vec.push_back(new VStore(src, gr2, 0));
                }
            } else {
                vec.push_back(new VStore(src, getGR(storeIr->dst), 0));
            }
        }
        return vec;
    }
    if (typeid(*ir) == typeid(UnaryIR)) {
        std::vector<Instr *> vec;
        UnaryIR *unaryIr = dynamic_cast<UnaryIR *>(ir);
        if (unaryIr->res.isInt()) {
            if (unaryIr->op == OP::NEG) {
                vec.push_back(new RsubImm(getGR(unaryIr->res.getVal()), getGR(unaryIr->v.getVal()), 0));
            } else {
                vec.push_back(new CmpImm(getGR(unaryIr->v.getVal()), 0));
                vec.push_back(new MovImm(getGR(unaryIr->res.getVal()), 0));
                vec.push_back(new MovImm(getGR(unaryIr->res.getVal()), 1, EQU));
            }
        } else {
            if (unaryIr->op == OP::NEG) {
                vec.push_back(new VNeg(getFR(unaryIr->res.getVal()), getFR(unaryIr->v.getVal())));
            } else {
                vec.push_back(new VCmpe(getFR(unaryIr->v.getVal()), 0));
                vec.push_back(new VMrs());
                GR dst = getGR(unaryIr->res.getVal());
                vec.push_back(new MovImm(dst, 0));
                vec.push_back(new MovImm(dst, 1, EQU));
            }
        }
        return vec;
    }
    if (typeid(*ir) == typeid(CastInt2FloatIR)) {
        CastInt2FloatIR *ir2 = dynamic_cast<CastInt2FloatIR *>(ir);
        GR src = getGR(ir2->v2);
        FR temp = FR::allocateReg();
        FR dst = getFR(ir2->v1);
        return {new VMovFG(temp, src), new VcvtFS(dst, temp)};
    }
    if (typeid(*ir) == typeid(CastFloat2IntIR)) {
        CastFloat2IntIR *ir2 = dynamic_cast<CastFloat2IntIR *>(ir);
        FR src = getFR(ir2->v2);
        FR temp = FR::allocateReg();
        GR dst = getGR(ir2->v1);
        return {new VcvtSF(temp, src), new VMovGF(dst, temp)};
    }

    if (typeid(*ir) == typeid(AddIIR)) {
        std::vector<Instr *> vec;
        AddIIR *addIir = dynamic_cast<AddIIR *>(ir);
        if (!addIir->left.getVal() && addIir->right.getVal()) {
            TempVal temp = addIir->left;
            addIir->left = addIir->right;
            addIir->right = temp;
        }
        if (!addIir->right.getVal()) {
            if (is_legal_immediate(addIir->right.getInt())) {
                vec.push_back(
                        new GRegImmInstr(GRegImmInstr::Add, getGR(addIir->res.getVal()), getGR(addIir->left.getVal()),
                                         addIir->right.getInt()));
            } else if (is_legal_immediate(-addIir->right.getInt())) {
                vec.push_back(
                        new GRegImmInstr(GRegImmInstr::Sub, getGR(addIir->res.getVal()), getGR(addIir->left.getVal()),
                                         -addIir->right.getInt()));
            } else {
                for (Instr *instr: setIntValue(GR(12), addIir->right.getInt())) {
                    vec.push_back(instr);
                }
                vec.push_back(new GRegRegInstr(GRegRegInstr::Add,
                                               getGR(addIir->res.getVal()),
                                               getGR(addIir->left.getVal()),
                                               GR(12)));
            }
        } else {
            if (gRegMapping.count(addIir->right.getVal()) == 0) {
                std::cerr << "error!\n";
            }
            vec.push_back(new GRegRegInstr(GRegRegInstr::Add, getGR(addIir->res.getVal()), getGR(addIir->left.getVal()),
                                     getGR(addIir->right.getVal())));


        }
        return vec;
    }
    if (typeid(*ir) == typeid(SubIIR)) {
        std::vector<Instr *> vec;
        SubIIR *subIir = dynamic_cast<SubIIR *>(ir);
        if (!subIir->right.getVal()) {
            if (is_legal_immediate(subIir->right.getInt())) {
                vec.push_back(
                        new GRegImmInstr(GRegImmInstr::Sub, getGR(subIir->res.getVal()), getGR(subIir->left.getVal()),
                                         subIir->right.getInt()));
            } else if (is_legal_immediate(-subIir->right.getInt())) {
                vec.push_back(
                        new GRegImmInstr(GRegImmInstr::Add, getGR(subIir->res.getVal()), getGR(subIir->left.getVal()),
                                         -subIir->right.getInt()));
            } else {
                for (Instr *instr: setIntValue(GR(12), subIir->right.getInt())) {
                    vec.push_back(instr);
                }
                vec.push_back(new GRegRegInstr(GRegRegInstr::Sub,
                                               getGR(subIir->res.getVal()),
                                               getGR(subIir->left.getVal()),
                                               GR(12)));
            }
        } else if (!subIir->left.getVal()) {
            if (is_legal_immediate(subIir->left.getInt())) {
                vec.push_back(
                        new GRegImmInstr(GRegImmInstr::RSUB, getGR(subIir->res.getVal()), getGR(subIir->right.getVal()),
                                         subIir->left.getInt()));
            } else {
                for (Instr *instr: setIntValue(GR(12), subIir->left.getInt())) {
                    vec.push_back(instr);
                }
                vec.push_back(new GRegRegInstr(GRegRegInstr::Sub,
                                               getGR(subIir->res.getVal()),
                                               GR(12),
                                               getGR(subIir->right.getVal())));
            }
        } else {
            vec.push_back(new GRegRegInstr(GRegRegInstr::Sub, getGR(subIir->res.getVal()), getGR(subIir->left.getVal()),
                                           getGR(subIir->right.getVal())));
        }
        return vec;
    }
    if (typeid(*ir) == typeid(MulIIR)) {
        std::vector<Instr *> vec;
        MulIIR *mulIir = dynamic_cast<MulIIR *>(ir);
        if (!mulIir->left.getVal() && mulIir->right.getVal()) {
            TempVal temp = mulIir->left;
            mulIir->left = mulIir->right;
            mulIir->right = temp;
        }
        if (!mulIir->right.getVal()) {
            int v = mulIir->right.getInt();
            bool flag = false;
            //add inst * 1
            {
                for (int lsl_1 = 0; lsl_1 < 32; lsl_1++) {
                    int x = 1;
                    int y;
                    y = (x + (x << lsl_1));
                    if (y == v) {
                        flag = true;
                        vec.push_back(new GRegRegInstr(GRegRegInstr::Add, getGR(mulIir->res.getVal()),
                                                       getGR(mulIir->left.getVal()),
                                                       getGR(mulIir->left.getVal()), GRegRegInstr::LSL, lsl_1));
                        break;
                    }
                }
            }
            //sub inst*1
            {
                for (int lsl_1 = 0; lsl_1 < 32; lsl_1++) {
                    int x = 1;
                    int y;
                    y = (x - (x << lsl_1));
                    if (y == v) {
                        flag = true;
                        vec.push_back(new GRegRegInstr(GRegRegInstr::Sub, getGR(mulIir->res.getVal()),
                                                       getGR(mulIir->left.getVal()),
                                                       getGR(mulIir->left.getVal()), GRegRegInstr::LSL, lsl_1));
                        break;
                    }
                }
            }
            //rsb inst*1
            {
                for (int lsl_1 = 0; lsl_1 < 32; lsl_1++) {
                    int x = 1;
                    int y;
                    y = ((x << lsl_1) - x);
                    if (y == v) {
                        flag = true;
                        vec.push_back(new GRegRegInstr(GRegRegInstr::RSUB, getGR(mulIir->res.getVal()),
                                                       getGR(mulIir->left.getVal()),
                                                       getGR(mulIir->left.getVal()), GRegRegInstr::LSL, lsl_1));
                        break;
                    }
                }
            }
            //lsl inst*1
            {
                for (int lsl_1 = 0; lsl_1 < 32; lsl_1++) {
                    int x = 1;
                    int y;
                    y = (x << lsl_1);
                    if (y == v) {
                        flag = true;
                        vec.push_back(new LSImmInstr(LSImmInstr::LSL, getGR(mulIir->res.getVal()),
                                                     getGR(mulIir->left.getVal()), lsl_1));
                        break;
                    }
                }
            }
            if (!flag) {
                for (Instr *instr: setIntValue(GR(12), mulIir->right.getInt())) {
                    vec.push_back(instr);
                }
                vec.push_back(new GRegRegInstr(GRegRegInstr::Mul,
                                               getGR(mulIir->res.getVal()),
                                               getGR(mulIir->left.getVal()),
                                               GR(12)));
            }
        } else {
            vec.push_back(new GRegRegInstr(GRegRegInstr::Mul, getGR(mulIir->res.getVal()), getGR(mulIir->left.getVal()),
                                           getGR(mulIir->right.getVal())));
        }
        return vec;
    }
    if (typeid(*ir) == typeid(DivIIR)) {
        std::vector<Instr *> vec;
        DivIIR *divIir = dynamic_cast<DivIIR *>(ir);
        if (!divIir->right.getVal()) {
            int v = divIir->right.getInt();
            int i = 0;
            for (i = 0; i <= 8; ++i) {
                int x = 1;
                int y = x << i;
                if (y == v) {
                    GR temp = GR::allocateReg();
                    vec.push_back(new MoveReg(temp,
                                              getGR(divIir->left.getVal())));
                    vec.push_back(new CmpImm(temp, 0));
                    vec.push_back(new GRegImmInstr(GRegImmInstr::Add,
                                                   temp,
                                                   temp,
                                                   (1 << i) - 1, COND::LT));
                    vec.push_back(new MoveReg(getGR(divIir->res.getVal()), temp, i));
                    break;
                }
            }
            if (i == 9) {
                for (Instr *instr: setIntValue(GR(12), divIir->right.getInt())) {
                    vec.push_back(instr);
                }
                vec.push_back(new GRegRegInstr(GRegRegInstr::Div,
                                               getGR(divIir->res.getVal()),
                                               getGR(divIir->left.getVal()),
                                               GR(12)));
            }
        } else if (!divIir->left.getVal()) {
            if (!divIir->left.getVal()) {
                for (Instr *instr: setIntValue(GR(12), divIir->left.getInt())) {
                    vec.push_back(instr);
                }
                vec.push_back(new GRegRegInstr(GRegRegInstr::Div,
                                               getGR(divIir->res.getVal()),
                                               GR(12),
                                               getGR(divIir->right.getVal())));
            }
        } else {
            vec.push_back(new GRegRegInstr(GRegRegInstr::Div, getGR(divIir->res.getVal()), getGR(divIir->left.getVal()),
                                           getGR(divIir->right.getVal())));
        }
        return vec;
    }
    if (typeid(*ir) == typeid(ModIR)) {
        ModIR *modIr = dynamic_cast<ModIR *>(ir);
        std::vector<Instr *> vec;
        GR res_gr = getGR(modIr->res.getVal());
        GR left_gr, right_gr;
        if (modIr->left.getVal()) {
            left_gr = getGR(modIr->left.getVal());
        } else {
            for (Instr *instr: setIntValue(GR(12), modIr->left.getInt())) {
                vec.push_back(instr);
            }
            left_gr = GR(12);
        }
        if (modIr->right.getVal()) {
            right_gr = getGR(modIr->right.getVal());
        } else {
            for (Instr *instr: setIntValue(GR(12), modIr->right.getInt())) {
                vec.push_back(instr);
            }
            right_gr = GR(12);
        }
        vec.push_back(new GRegRegInstr(GRegRegInstr::Div, res_gr, left_gr, right_gr));
        vec.push_back(new GRegRegInstr(GRegRegInstr::Mul, res_gr, res_gr, right_gr));
        vec.push_back(new GRegRegInstr(GRegRegInstr::Sub, res_gr, left_gr, res_gr));
        return vec;
    }
    if (typeid(*ir) == typeid(LTIIR)) {
        LTIIR *ir2 = dynamic_cast<LTIIR *>(ir);
        std::vector<Instr *> vec;
        BranchIR* branchIr = dynamic_cast<BranchIR *>(*(iter + 1));
        iter++;
        assert(branchIr != nullptr);
        if (!ir2->right.getVal()) {
            if (is_legal_immediate(ir2->right.getInt())) {
                vec.push_back(new CmpImm(getGR(ir2->left.getVal()), ir2->right.getInt()));
            } else {
                for (Instr *instr: setIntValue(GR(12), ir2->right.getInt())) {
                    vec.push_back(instr);
                }
                GR right_gr = GR(12);
                vec.push_back(new Cmp(getGR(ir2->left.getVal()), right_gr));
            }
            vec.push_back(new B(branchIr->trueTarget->name, LT));
            vec.push_back(new B(branchIr->falseTarget->name));
//            vec.push_back(new MovImm(getGR(ir2->res.getVal()), 0));
//            vec.push_back(new MovImm(getGR(ir2->res.getVal()), 1, LT));
        } else if (!ir2->left.getVal()) {
            if (is_legal_immediate(ir2->left.getInt())) {
                vec.push_back(new CmpImm(getGR(ir2->right.getVal()), ir2->left.getInt()));
            } else {
                for (Instr *instr: setIntValue(GR(12), ir2->left.getInt())) {
                    vec.push_back(instr);
                }
                GR left_gr = GR(12);
                vec.push_back(new Cmp(getGR(ir2->right.getVal()), left_gr));
            }
            vec.push_back(new B(branchIr->trueTarget->name, GE));
            vec.push_back(new B(branchIr->falseTarget->name));
//            vec.push_back(new MovImm(getGR(ir2->res.getVal()), 0));
//            vec.push_back(new MovImm(getGR(ir2->res.getVal()), 1, GE));
        } else {
            vec.push_back(new Cmp(getGR(ir2->left.getVal()), getGR(ir2->right.getVal())));
            vec.push_back(new B(branchIr->trueTarget->name, LT));
            vec.push_back(new B(branchIr->falseTarget->name));
//            vec.push_back(new MovImm(getGR(ir2->res.getVal()), 0));
//            vec.push_back(new MovImm(getGR(ir2->res.getVal()), 1, LT));
        }
        return vec;
    }
    if (typeid(*ir) == typeid(LEIIR)) {
        LEIIR *ir2 = dynamic_cast<LEIIR *>(ir);
        std::vector<Instr *> vec;
        BranchIR* branchIr = dynamic_cast<BranchIR *>(*(iter + 1));
        iter++;
        assert(branchIr != nullptr);
        if (!ir2->right.getVal()) {
            if (is_legal_immediate(ir2->right.getInt())) {
                vec.push_back(new CmpImm(getGR(ir2->left.getVal()), ir2->right.getInt()));
            } else {
                for (Instr *instr: setIntValue(GR(12), ir2->right.getInt())) {
                    vec.push_back(instr);
                }
                GR right_gr = GR(12);
                vec.push_back(new Cmp(getGR(ir2->left.getVal()), right_gr));
            }
            vec.push_back(new B(branchIr->trueTarget->name, LE));
            vec.push_back(new B(branchIr->falseTarget->name));
//            vec.push_back(new MovImm(getGR(ir2->res.getVal()), 0));
//            vec.push_back(new MovImm(getGR(ir2->res.getVal()), 1, LE));
        } else if (!ir2->left.getVal()) {
            if (is_legal_immediate(ir2->left.getInt())) {
                vec.push_back(new CmpImm(getGR(ir2->right.getVal()), ir2->left.getInt()));
            } else {
                for (Instr *instr: setIntValue(GR(12), ir2->left.getInt())) {
                    vec.push_back(instr);
                }
                GR left_gr = GR(12);
                vec.push_back(new Cmp(getGR(ir2->right.getVal()), left_gr));
            }
            vec.push_back(new B(branchIr->trueTarget->name, GT));
            vec.push_back(new B(branchIr->falseTarget->name));
//            vec.push_back(new MovImm(getGR(ir2->res.getVal()), 0));
//            vec.push_back(new MovImm(getGR(ir2->res.getVal()), 1, GT));
        } else {
            vec.push_back(new Cmp(getGR(ir2->left.getVal()), getGR(ir2->right.getVal())));
            vec.push_back(new B(branchIr->trueTarget->name, LE));
            vec.push_back(new B(branchIr->falseTarget->name));
//            vec.push_back(new MovImm(getGR(ir2->res.getVal()), 0));
//            vec.push_back(new MovImm(getGR(ir2->res.getVal()), 1, LE));
        }
        return vec;
    }
    if (typeid(*ir) == typeid(GTIIR)) {
        GTIIR *ir2 = dynamic_cast<GTIIR *>(ir);
        std::vector<Instr *> vec;
        BranchIR* branchIr = dynamic_cast<BranchIR *>(*(iter + 1));
        iter++;
        assert(branchIr != nullptr);
        if (!ir2->right.getVal()) {
            if (is_legal_immediate(ir2->right.getInt())) {
                vec.push_back(new CmpImm(getGR(ir2->left.getVal()), ir2->right.getInt()));
            } else {
                for (Instr *instr: setIntValue(GR(12), ir2->right.getInt())) {
                    vec.push_back(instr);
                }
                GR right_gr = GR(12);
                vec.push_back(new Cmp(getGR(ir2->left.getVal()), right_gr));
            }
            vec.push_back(new B(branchIr->trueTarget->name, GT));
            vec.push_back(new B(branchIr->falseTarget->name));
//            vec.push_back(new MovImm(getGR(ir2->res.getVal()), 0));
//            vec.push_back(new MovImm(getGR(ir2->res.getVal()), 1, GT));
        } else if (!ir2->left.getVal()) {
            if (is_legal_immediate(ir2->left.getInt())) {
                vec.push_back(new CmpImm(getGR(ir2->right.getVal()), ir2->left.getInt()));
            } else {
                for (Instr *instr: setIntValue(GR(12), ir2->left.getInt())) {
                    vec.push_back(instr);
                }
                GR left_gr = GR(12);
                vec.push_back(new Cmp(getGR(ir2->right.getVal()), left_gr));
            }
            vec.push_back(new B(branchIr->trueTarget->name, LE));
            vec.push_back(new B(branchIr->falseTarget->name));
//            vec.push_back(new MovImm(getGR(ir2->res.getVal()), 0));
//            vec.push_back(new MovImm(getGR(ir2->res.getVal()), 1, LE));
        } else {
            vec.push_back(new Cmp(getGR(ir2->left.getVal()), getGR(ir2->right.getVal())));
            vec.push_back(new B(branchIr->trueTarget->name, GT));
            vec.push_back(new B(branchIr->falseTarget->name));
//            vec.push_back(new MovImm(getGR(ir2->res.getVal()), 0));
//            vec.push_back(new MovImm(getGR(ir2->res.getVal()), 1, GT));
        }
        return vec;
    }
    if (typeid(*ir) == typeid(GEIIR)) {
        GEIIR *ir2 = dynamic_cast<GEIIR *>(ir);
        std::vector<Instr *> vec;
        BranchIR* branchIr = dynamic_cast<BranchIR *>(*(iter + 1));
        iter++;
        assert(branchIr != nullptr);
        if (!ir2->right.getVal()) {
            if (is_legal_immediate(ir2->right.getInt())) {
                vec.push_back(new CmpImm(getGR(ir2->left.getVal()), ir2->right.getInt()));
            } else {
                for (Instr *instr: setIntValue(GR(12), ir2->right.getInt())) {
                    vec.push_back(instr);
                }
                GR right_gr = GR(12);
                vec.push_back(new Cmp(getGR(ir2->left.getVal()), right_gr));
            }
            vec.push_back(new B(branchIr->trueTarget->name, GE));
            vec.push_back(new B(branchIr->falseTarget->name));
//            vec.push_back(new MovImm(getGR(ir2->res.getVal()), 0));
//            vec.push_back(new MovImm(getGR(ir2->res.getVal()), 1, GE));
        } else if (!ir2->left.getVal()) {
            if (is_legal_immediate(ir2->left.getInt())) {
                vec.push_back(new CmpImm(getGR(ir2->right.getVal()), ir2->left.getInt()));
            } else {
                for (Instr *instr: setIntValue(GR(12), ir2->left.getInt())) {
                    vec.push_back(instr);
                }
                GR left_gr = GR(12);
                vec.push_back(new Cmp(getGR(ir2->right.getVal()), left_gr));
            }
            vec.push_back(new B(branchIr->trueTarget->name, LT));
            vec.push_back(new B(branchIr->falseTarget->name));
//            vec.push_back(new MovImm(getGR(ir2->res.getVal()), 0));
//            vec.push_back(new MovImm(getGR(ir2->res.getVal()), 1, LT));
        } else {
            vec.push_back(new Cmp(getGR(ir2->left.getVal()), getGR(ir2->right.getVal())));
            vec.push_back(new B(branchIr->trueTarget->name, GE));
            vec.push_back(new B(branchIr->falseTarget->name));
//            vec.push_back(new MovImm(getGR(ir2->res.getVal()), 0));
//            vec.push_back(new MovImm(getGR(ir2->res.getVal()), 1, GE));
        }
        return vec;
    }
    if (typeid(*ir) == typeid(EQUIIR)) {
        EQUIIR *ir2 = dynamic_cast<EQUIIR *>(ir);
        std::vector<Instr *> vec;
        BranchIR* branchIr = dynamic_cast<BranchIR *>(*(iter + 1));
        iter++;
        assert(branchIr != nullptr);
        if (!ir2->right.getVal()) {
            if (is_legal_immediate(ir2->right.getInt())) {
                vec.push_back(new CmpImm(getGR(ir2->left.getVal()), ir2->right.getInt()));
            } else {
                for (Instr *instr: setIntValue(GR(12), ir2->right.getInt())) {
                    vec.push_back(instr);
                }
                GR right_gr = GR(12);
                vec.push_back(new Cmp(getGR(ir2->left.getVal()), right_gr));
            }
            vec.push_back(new B(branchIr->trueTarget->name, EQU));
            vec.push_back(new B(branchIr->falseTarget->name));
//            vec.push_back(new MovImm(getGR(ir2->res.getVal()), 0));
//            vec.push_back(new MovImm(getGR(ir2->res.getVal()), 1, EQU));
        } else if (!ir2->left.getVal()) {
            if (is_legal_immediate(ir2->left.getInt())) {
                vec.push_back(new CmpImm(getGR(ir2->right.getVal()), ir2->left.getInt()));
            } else {
                for (Instr *instr: setIntValue(GR(12), ir2->left.getInt())) {
                    vec.push_back(instr);
                }
                GR left_gr = GR(12);
                vec.push_back(new Cmp(getGR(ir2->right.getVal()), left_gr));
            }
            vec.push_back(new B(branchIr->trueTarget->name, EQU));
            vec.push_back(new B(branchIr->falseTarget->name));
//            vec.push_back(new MovImm(getGR(ir2->res.getVal()), 0));
//            vec.push_back(new MovImm(getGR(ir2->res.getVal()), 1, EQU));
        } else {
            vec.push_back(new Cmp(getGR(ir2->left.getVal()), getGR(ir2->right.getVal())));
            vec.push_back(new B(branchIr->trueTarget->name, EQU));
            vec.push_back(new B(branchIr->falseTarget->name));
//            vec.push_back(new MovImm(getGR(ir2->res.getVal()), 0));
//            vec.push_back(new MovImm(getGR(ir2->res.getVal()), 1, EQU));
        }
        return vec;
    }
    if (typeid(*ir) == typeid(NEIIR)) {
        NEIIR *ir2 = dynamic_cast<NEIIR *>(ir);
        std::vector<Instr *> vec;
        BranchIR* branchIr = dynamic_cast<BranchIR *>(*(iter + 1));
        iter++;
        assert(branchIr != nullptr);
        if (!ir2->right.getVal()) {
            if (is_legal_immediate(ir2->right.getInt())) {
                vec.push_back(new CmpImm(getGR(ir2->left.getVal()), ir2->right.getInt()));
            } else {
                for (Instr *instr: setIntValue(GR(12), ir2->right.getInt())) {
                    vec.push_back(instr);
                }
                GR right_gr = GR(12);
                vec.push_back(new Cmp(getGR(ir2->left.getVal()), right_gr));
            }
            vec.push_back(new B(branchIr->trueTarget->name, NE));
            vec.push_back(new B(branchIr->falseTarget->name));
//            vec.push_back(new MovImm(getGR(ir2->res.getVal()), 0));
//            vec.push_back(new MovImm(getGR(ir2->res.getVal()), 1, NE));
        } else if (!ir2->left.getVal()) {
            if (is_legal_immediate(ir2->left.getInt())) {
                vec.push_back(new CmpImm(getGR(ir2->right.getVal()), ir2->left.getInt()));
            } else {
                for (Instr *instr: setIntValue(GR(12), ir2->left.getInt())) {
                    vec.push_back(instr);
                }
                GR left_gr = GR(12);
                vec.push_back(new Cmp(getGR(ir2->right.getVal()), left_gr));
            }
            vec.push_back(new B(branchIr->trueTarget->name, NE));
            vec.push_back(new B(branchIr->falseTarget->name));
//            vec.push_back(new MovImm(getGR(ir2->res.getVal()), 0));
//            vec.push_back(new MovImm(getGR(ir2->res.getVal()), 1, NE));
        } else {
            vec.push_back(new Cmp(getGR(ir2->left.getVal()), getGR(ir2->right.getVal())));
            vec.push_back(new B(branchIr->trueTarget->name, NE));
            vec.push_back(new B(branchIr->falseTarget->name));
//            vec.push_back(new MovImm(getGR(ir2->res.getVal()), 0));
//            vec.push_back(new MovImm(getGR(ir2->res.getVal()), 1, NE));
        }
        return vec;
    }

    if (typeid(*ir) == typeid(AddFIR) || typeid(*ir) == typeid(SubFIR) ||
        typeid(*ir) == typeid(MulFIR) || typeid(*ir) == typeid(DivFIR) ||
        typeid(*ir) == typeid(LTFIR) || typeid(*ir) == typeid(GTFIR) ||
        typeid(*ir) == typeid(LEFIR) || typeid(*ir) == typeid(GEFIR) ||
        typeid(*ir) == typeid(EQUFIR) || typeid(*ir) == typeid(NEFIR)) {
        ArithmeticIR *ir2 = dynamic_cast<ArithmeticIR *>(ir);
        std::vector<Instr *> vec;
        FR res_fr = getFR(ir2->res.getVal());
        FR left_fr, right_fr;
        if (!ir2->left.getVal()) {
            left_fr = getFloatFR(ir2->left.getFloat(), block, vec);
        } else {
            left_fr = getFR(ir2->left.getVal());
        }
        if (!ir2->right.getVal()) {
            right_fr = getFloatFR(ir2->right.getFloat(), block, vec);
        } else {
            right_fr = getFR(ir2->right.getVal());
        }
        if (typeid(*ir) == typeid(AddFIR)) {
            vec.push_back(new VRegRegInstr(VRegRegInstr::VAdd, res_fr, left_fr, right_fr));
        }
        if (typeid(*ir) == typeid(SubFIR)) {
            vec.push_back(new VRegRegInstr(VRegRegInstr::VSub, res_fr, left_fr, right_fr));
        }
        if (typeid(*ir) == typeid(MulFIR)) {
            vec.push_back(new VRegRegInstr(VRegRegInstr::VMul, res_fr, left_fr, right_fr));
        }
        if (typeid(*ir) == typeid(DivFIR)) {
            vec.push_back(new VRegRegInstr(VRegRegInstr::VDiv, res_fr, left_fr, right_fr));
        }
        if (typeid(*ir) == typeid(LTFIR)) {
            vec.push_back(new VCmpe(left_fr, right_fr));
            vec.push_back(new VMrs());
            vec.push_back(new MovImm(getGR(ir2->res.getVal()), 0));
            vec.push_back(new MovImm(getGR(ir2->res.getVal()), 1, LT));
        }
        if (typeid(*ir) == typeid(GTFIR)) {
            vec.push_back(new VCmpe(left_fr, right_fr));
            vec.push_back(new VMrs());
            vec.push_back(new MovImm(getGR(ir2->res.getVal()), 0));
            vec.push_back(new MovImm(getGR(ir2->res.getVal()), 1, GT));
        }
        if (typeid(*ir) == typeid(LEFIR)) {
            vec.push_back(new VCmpe(left_fr, right_fr));
            vec.push_back(new VMrs());
            vec.push_back(new MovImm(getGR(ir2->res.getVal()), 0));
            vec.push_back(new MovImm(getGR(ir2->res.getVal()), 1, LE));
        }
        if (typeid(*ir) == typeid(GEFIR)) {
            vec.push_back(new VCmpe(left_fr, right_fr));
            vec.push_back(new VMrs());
            vec.push_back(new MovImm(getGR(ir2->res.getVal()), 0));
            vec.push_back(new MovImm(getGR(ir2->res.getVal()), 1, GE));
        }
        if (typeid(*ir) == typeid(EQUFIR)) {
            vec.push_back(new VCmpe(left_fr, right_fr));
            vec.push_back(new VMrs());
            vec.push_back(new MovImm(getGR(ir2->res.getVal()), 0));
            vec.push_back(new MovImm(getGR(ir2->res.getVal()), 1, EQU));
        }
        if (typeid(*ir) == typeid(NEFIR)) {
            vec.push_back(new VCmpe(left_fr, right_fr));
            vec.push_back(new VMrs());
            vec.push_back(new MovImm(getGR(ir2->res.getVal()), 0));
            vec.push_back(new MovImm(getGR(ir2->res.getVal()), 1, NE));
        }
        return vec;
    }
    if (typeid(*ir) == typeid(JumpIR)) {
        return {new B(dynamic_cast<JumpIR *>(ir)->target->name)};
    }
    if (typeid(*ir) == typeid(BranchIR)) {
        BranchIR *branchIr = dynamic_cast<BranchIR *>(ir);
        return {new CmpImm(getGR(branchIr->cond), 0),
                new B(branchIr->trueTarget->name, NE),
                new B(branchIr->falseTarget->name, EQU)};
    }
    if (typeid(*ir) == typeid(CallIR)) {
        CallIR *callIr = dynamic_cast<CallIR *>(ir);
        int gr_cnt = 0;
        int fr_cnt = 0;
        int cnt = 0;
        std::vector<Instr *> vec;
        for (TempVal v: callIr->args) {
            std::vector<Instr *> tempVec;
            if (v.getType()->isString()) {
                tempVec.push_back(new MoveWFromSymbol(GR(0), stringConstMapping[v.getString()]));
                tempVec.push_back(new MoveTFromSymbol(GR(0), stringConstMapping[v.getString()]));
                gr_cnt++;
                continue;
            }
            if (!v.isFloat()) {
                if (gr_cnt >= 4) {
                    if (!v.getVal()) {
                        for (Instr *instr: setIntValue(GR(12), v.getInt())) {
                            tempVec.push_back(instr);
                        }
                        GR gr = GR(12);
                        tempVec.push_back(new Store(gr, GR(13), cnt * 4));
                    } else if (gRegMapping.count(v.getVal()) == 0) {
                        if (is_legal_load_store_offset(stackMapping[v.getVal()]) &&
                            is_legal_immediate(stackMapping[v.getVal()])) {
                            GR gr = GR::allocateReg();
                            tempVec.push_back(new GRegImmInstr(GRegImmInstr::Add, gr, GR(13),
                                                               stackMapping[v.getVal()]));
                            tempVec.push_back(new Store(gr, GR(13), cnt * 4));
                        } else {
                            GR gr = GR::allocateReg();
                            for (Instr *instr: setIntValue(GR(12), stackMapping[v.getVal()])) {
                                tempVec.push_back(instr);
                            }
                            GR constant = GR(12);
                            tempVec.push_back(new GRegRegInstr(GRegRegInstr::Add, gr, GR(13), constant));
                            tempVec.push_back(new Store(gr, GR(13), cnt * 4));
                        }
                    } else {
                        tempVec.push_back(new Store(getGR(v.getVal()), GR(13), cnt * 4));
                    }
                    cnt++;
                } else {
                    if (!v.getVal()) {
                        if (is_legal_load_store_offset(v.getInt()) && is_legal_immediate(v.getInt())) {
                            tempVec.push_back(new MovImm(GR(gr_cnt), v.getInt()));
                        } else {
                            for (Instr *instr: setIntValue(GR(12), v.getInt())) {
                                tempVec.push_back(instr);
                            }
                            GR gr = GR(12);
                            tempVec.push_back(new MoveReg(GR(gr_cnt), gr));
                        }
                    } else {
                        if (gRegMapping.count(v.getVal()) != 0) {
                            tempVec.push_back(new MoveReg(GR(gr_cnt), getGR(v.getVal())));
                        } else {
                            if (is_legal_load_store_offset(stackMapping[v.getVal()]) &&
                                is_legal_immediate(stackMapping[v.getVal()])) {
                                tempVec.push_back(new GRegImmInstr(GRegImmInstr::Add, GR(gr_cnt), GR(13),
                                                                   stackMapping[v.getVal()]));
                            } else {
                                for (Instr *instr: setIntValue(GR(12), stackMapping[v.getVal()])) {
                                    vec.push_back(instr);
                                }
                                GR gr = GR(12);
                                tempVec.push_back(new GRegRegInstr(GRegRegInstr::Add, GR(gr_cnt), GR(13), gr));
                            }
                        }
                    }
                }
                gr_cnt++;
            } else {
                if (fr_cnt >= 32) {
                    if (!v.getVal()) {
                        FR fr = getFloatFR(v.getFloat(), block, vec);
                    }
                    tempVec.push_back(new VStore(getFR(v.getVal()), GR(13), cnt * 4));
                    cnt++;
                } else {
                    if (!v.getVal()) {
                        if (constantFloatMapping[block].count(v.getFloat()) != 0) {
                            tempVec.push_back(new VMoveReg(FR(fr_cnt), constantFloatMapping[block][v.getFloat()]));
                        } else {
                            tempVec.push_back(new MoveWFromSymbol(GR(12), getFloatAddr(v.getFloat())));
                            tempVec.push_back(new MoveTFromSymbol(GR(12), getFloatAddr(v.getFloat())));
                            tempVec.push_back(new VLoad(FR(fr_cnt), GR(12), 0));
                        }
                    } else {
                        tempVec.push_back(new VMoveReg(FR(fr_cnt), getFR(v.getVal())));
                    }
                }
                fr_cnt++;
            }
            for (auto it = tempVec.begin(); it != tempVec.end(); it++) {
                //vec.insert(vec.begin(), *it);
                vec.push_back(*it);
            }
        }

//        vec.push_back(new Push({}));
//        vec.push_back(new MoveReg(GR(13),GR(11)));
        vec.push_back(new Bl(callIr->func->name));
//        vec.push_back(new MoveReg(GR(11),GR(13)));
//        vec.push_back(new Pop({}));
        if (callIr->returnVal) {
            if (callIr->returnVal->getType()->isInt()) {
                vec.push_back(new MoveReg(getGR(callIr->returnVal), GR(0)));
            } else {
                vec.push_back(new VMoveReg(getFR(callIr->returnVal), FR(0)));
            }
        }
        return vec;
    }
    if (typeid(*ir) == typeid(ReturnIR)) {
        ReturnIR *returnIr = dynamic_cast<ReturnIR *>(ir);
        std::vector<Instr *> vec;
        if (returnIr->useInt) {
            vec.push_back(new MovImm(GR(0), returnIr->retInt));
        } else if (returnIr->useFloat) {
            if (constantFloatMapping[block].count(returnIr->retFloat) != 0) {
                vec.push_back(new VMoveReg(FR(0), constantFloatMapping[block][returnIr->retFloat]));
            } else {
                vec.push_back(new MoveWFromSymbol(GR(12), getFloatAddr(returnIr->retFloat)));
                vec.push_back(new MoveTFromSymbol(GR(12), getFloatAddr(returnIr->retFloat)));
                vec.push_back(new VLoad(FR(0), GR(12), 0));
            }
        } else if (returnIr->v) {
            if (returnIr->v->getType()->isInt()) {
                vec.push_back(new MoveReg(GR(0), getGR(returnIr->v)));
            } else {
                vec.push_back(new VMoveReg(FR(0), getFR(returnIr->v)));
            }
        }
        vec.push_back(new Ret());
        return vec;
    }
    return {};
}

GR Codegen::getGR(Value *src) {
    if (gRegMapping.count(src) == 0) {
        GR reg = GR::allocateReg();
        gRegMapping[src] = reg;
        return reg;
    }
    return gRegMapping[src];
}

FR Codegen::getFR(Value *src) {
    if (fRegMapping.count(src) == 0) {
        FR reg = FR::allocateReg();
        fRegMapping[src] = reg;
        return reg;
    }
    return fRegMapping[src];
}

void Codegen::generateGlobalCode() {
    std::vector<Value *> dataList;
    std::vector<Value *> bssList;
    for (Value *v: irVisitor.globalVars) {
        if (typeid(*v) == typeid(ConstValue)) {
            dataList.push_back(v);
        } else {
            if (v->getType()->isString()) {
                dataList.push_back(v);
            } else {
                bssList.push_back(v);
            }
        }
    }
    if (!dataList.empty()) {
        out << ".section .data\n";
        for (Value *v: dataList) {
            if (!v->getType()->isString()) {
                ConstValue *vv = dynamic_cast<ConstValue *>(v);
                if (!vv->is_Array() && !vv->getType()->isInt()) {
                    if (floatConstMapping.count(vv->getFloatVal()) == 0) {
                        floatConstMapping[vv->getFloatVal()] = vv->getName();
                    }
                    continue;
                }
                out << ".align\n";
                out << v->getName() << ":\n";
                out << "\t.4byte ";
                if (!vv->is_Array()) {
                    if (vv->getType()->isInt()) {
                        out << vv->getIntVal();
                    }
                } else {
                    for (int i = 0; i < vv->getArrayLen(); ++i) {
                        if (i != 0) {
                            out << ",";
                        }
                        if (vv->getType()->isIntPointer()) {
                            if (vv->getIntValList().count(i) == 0) {
                                out << "0";
                            } else {
                                out << vv->getIntValList()[i];
                            }
                        } else {
                            if (vv->getFloatValList().count(i) == 0) {
                                out << "0";
                            } else {
                                out << floatToInt(vv->getFloatValList()[i]);
                            }
                        }
                    }
                }
                out << "\n";
            } else {
                std::string varName = "_m_global_string_const" + std::to_string(stringConstCnt++);
                stringConstMapping[v->getName()] = varName;
                out << varName << ":\n\t.asciz " << v->getName() << "\n";
            }
        }
    }
    if (!bssList.empty()) {
        out << ".section .bss\n";
        for (Value *v: bssList) {
            out << ".align\n";
            out << v->getName() << ":\n";
            out << "\t.space ";
            if (!v->is_Array()) {
                out << "4";
            } else {
                out << v->getArrayLen() * 4;
            }
            out << "\n";
        }
    }
}

void Codegen::comment(std::string s) {
    out << "@ " << s << "\n";
}